"""Discover myStrom devices."""
from . import MDNSDiscoverable


# pylint: disable=too-few-public-methods
class Discoverable(MDNSDiscoverable):
    """Add support for discovering myStrom switches."""

    def __init__(self, nd):
        super(Discoverable, self).__init__(nd, '_hap._tcp.local.')



"""Discover Home Assistant servers."""
from . import MDNSDiscoverable


# pylint: disable=too-few-public-methods
class Discoverable(MDNSDiscoverable):
    """Add support for discovering Home Assistant instances."""

    def __init__(self, nd):
        super(Discoverable, self).__init__(nd, '_home-assistant._tcp.local.')



"""Command line tool to print discocvered devices or dump raw data."""
from pprint import pprint
import sys

from netdisco.discovery import NetworkDiscovery


def main():
    """Handle command line execution."""
    netdisco = NetworkDiscovery()

    netdisco.scan()

    print("Discovered devices:")
    count = 0
    for dev in netdisco.discover():
        count += 1
        print('{}:'.format(dev))
        pprint(netdisco.get_info(dev))
        print()
    print("Discovered {} devices".format(count))

    # Pass in command line argument dump to get the raw data
    if sys.argv[-1] == 'dump':
        print()
        print()
        print("Raw Data")
        print()
        netdisco.print_raw_data()

    netdisco.stop()


if __name__ == '__main__':
    main()



"""Discover Belkin Wemo devices."""
from . import SSDPDiscoverable
from ..const import ATTR_MAC_ADDRESS


class Discoverable(SSDPDiscoverable):
    """Add support for discovering Belkin WeMo platform devices."""

    def info_from_entry(self, entry):
        """Return most important info from a uPnP entry."""
        info = super().info_from_entry(entry)
        device = entry.description['device']
        info[ATTR_MAC_ADDRESS] = device.get('macAddress', '')
        return info

    def get_entries(self):
        """Return all Belkin Wemo entries."""
        return self.find_by_device_description(
            {'manufacturer': 'Belkin International Inc.'})



"""Philips Hue bridge discovery using N-UPnP.

Philips Hue bridge limits how many SSDP lookups you can make. To work
around this they recommend to do N-UPnP lookup simultaneously with
SSDP lookup: https://developers.meethue.com/documentation/hue-bridge-discovery

"""

import xml.etree.ElementTree as ElementTree

import logging
import requests

from netdisco.util import etree_to_dict

_LOGGER = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class PHueBridge(object):
    """Parses Philips Hue bridge description XML into an
    object similar to UPNPEntry.

    """

    def __init__(self, location, description_xml):
        self.location = location
        tree = ElementTree.fromstring(description_xml)
        self.description = etree_to_dict(tree).get("root", {})

    def __repr__(self):
        friendly_name = self.description['device']['friendlyName']
        url_base = self.description['URLBase']

        return str((friendly_name, url_base))


class PHueNUPnPDiscovery(object):
    """Philips Hue bridge discovery using N-UPnP."""

    PHUE_NUPNP_URL = "https://www.meethue.com/api/nupnp"
    DESCRIPTION_URL_TMPL = "http://{}/description.xml"

    def __init__(self):
        self.entries = []

    def scan(self):
        """Scan the network."""
        try:
            response = requests.get(self.PHUE_NUPNP_URL, timeout=5)
            response.raise_for_status()
            self.entries = []
            bridges = response.json()
            for bridge in bridges:
                entry = self.fetch_description(bridge)
                if entry:
                    self.entries.append(entry)
        except requests.exceptions.RequestException as err:
            _LOGGER.warning('Could not query server %s: %s',
                            self.PHUE_NUPNP_URL,
                            err)

    def fetch_description(self, bridge):
        """Fetches description XML of a Philips Hue bridge."""
        url = self.bridge_description_url(bridge)
        try:
            response = requests.get(url, timeout=5)
            response.raise_for_status()
            return PHueBridge(url, response.text)
        except requests.exceptions.RequestException as err:
            _LOGGER.warning('Could not query server %s: %s',
                            url, err)

    def bridge_description_url(self, bridge):
        """Returns URL for fetching description XML"""
        ipaddr = bridge["internalipaddress"]
        return self.DESCRIPTION_URL_TMPL.format(ipaddr)


def main():
    """Test N-UPnP discovery."""
    from pprint import pprint

    disco = PHueNUPnPDiscovery()
    disco.scan()
    pprint(disco.entries)


if __name__ == "__main__":
    main()



"""Discover Openhome devices."""
from . import SSDPDiscoverable


# pylint: disable=too-few-public-methods
class Discoverable(SSDPDiscoverable):
    """Add support for discovering Openhome compliant devices."""

    def get_entries(self):
        """Get all the Openhome compliant device uPnP entries."""
        return self.find_by_st("urn:av-openhome-org:service:Product:2")



"""Discover PlexMediaServer."""
from . import GDMDiscoverable
from ..const import ATTR_NAME, ATTR_HOST, ATTR_PORT, ATTR_URLBASE


class Discoverable(GDMDiscoverable):
    """Add support for discovering Plex Media Server."""

    def info_from_entry(self, entry):
        """Return most important info from a GDM entry."""
        return {
            ATTR_NAME: entry['data']['Name'],
            ATTR_HOST: entry['from'][0],
            ATTR_PORT: entry['data']['Port'],
            ATTR_URLBASE: 'https://%s:%s' % (entry['from'][0],
                                             entry['data']['Port'])
        }

    def get_entries(self):
        """Return all PMS entries."""
        return self.find_by_data({'Content-Type': 'plex/media-server'})



"""Discover DLNA services."""
from . import SSDPDiscoverable


# pylint: disable=too-few-public-methods
class Discoverable(SSDPDiscoverable):
    """Add support for discovering DLNA services."""

    def get_entries(self):
        """Get all the DLNA service uPnP entries."""
        return self.find_by_st("urn:schemas-upnp-org:device:MediaServer:1")



"""Discover Philips Hue bridges."""
from . import SSDPDiscoverable


class Discoverable(SSDPDiscoverable):
    """Add support for discovering Philips Hue bridges."""

    def get_entries(self):
        """Get all the Hue bridge uPnP entries."""
        nupnp_entries = self.netdis.phue.entries

        # Hub models for year 2012 and 2015
        ssdp_entries = self.find_by_device_description({
            "manufacturer": "Royal Philips Electronics",
            "modelNumber": ["929000226503", "BSB002"]
        })

        return self.merge_entries(nupnp_entries, ssdp_entries)

    # pylint: disable=no-self-use
    def merge_entries(self, nupnp_entries, ssdp_entries):
        """Takes lists of device entries found using N-UPnP & SSDP lookups and
        merges them making sure that same device is only discoverd once.

        """
        entries = {
            nupn_entry.description['URLBase']: nupn_entry
            for nupn_entry in nupnp_entries
        }

        for ssdp_entry in ssdp_entries:
            url_base = ssdp_entry.description['URLBase']
            if url_base not in entries:
                entries[url_base] = ssdp_entry

        return list(entries.values())



"""Discover Kodi servers."""
from . import MDNSDiscoverable


# pylint: disable=too-few-public-methods
class Discoverable(MDNSDiscoverable):
    """Add support for discovering Kodi."""

    def __init__(self, nd):
        """Initialize the Kodi discovery."""
        super(Discoverable, self).__init__(nd, '_http._tcp.local.')

    def get_entries(self):
        return self.find_by_device_name('Kodi ')



"""Module that implements SSDP protocol."""
import re
import select
import socket
import logging
from datetime import datetime, timedelta
import xml.etree.ElementTree as ElementTree

import requests

from netdisco.util import etree_to_dict, interface_addresses

DISCOVER_TIMEOUT = 5
SSDP_MX = 1
SSDP_TARGET = ("239.255.255.250", 1900)

RESPONSE_REGEX = re.compile(r'\n(.*)\: (.*)\r')

MIN_TIME_BETWEEN_SCANS = timedelta(seconds=59)

# Devices and services
ST_ALL = "ssdp:all"

# Devices only, some devices will only respond to this query
ST_ROOTDEVICE = "upnp:rootdevice"


class SSDP(object):
    """Control the scanning of uPnP devices and services and caches output."""

    def __init__(self):
        """Initialize the discovery."""
        self.entries = []
        self.last_scan = None

    def scan(self):
        """Scan the network."""
        self.update()

    def all(self):
        """Return all found entries.

        Will scan for entries if not scanned recently.
        """
        self.update()

        return list(self.entries)

    # pylint: disable=invalid-name
    def find_by_st(self, st):
        """Return a list of entries that match the ST."""
        self.update()

        return [entry for entry in self.entries
                if entry.st == st]

    def find_by_device_description(self, values):
        """Return a list of entries that match the description.

        Pass in a dict with values to match against the device tag in the
        description.
        """
        self.update()

        return [entry for entry in self.entries
                if entry.match_device_description(values)]

    def update(self, force_update=False):
        """Scan for new uPnP devices and services."""
        if self.last_scan is None or force_update or \
           datetime.now()-self.last_scan > MIN_TIME_BETWEEN_SCANS:

            self.remove_expired()

            self.entries.extend(
                entry for entry in scan()
                if entry not in self.entries)

            self.last_scan = datetime.now()

    def remove_expired(self):
        """Filter out expired entries."""
        self.entries = [entry for entry in self.entries
                        if not entry.is_expired]


class UPNPEntry(object):
    """Found uPnP entry."""

    DESCRIPTION_CACHE = {'_NO_LOCATION': {}}

    def __init__(self, values):
        """Initialize the discovery."""
        self.values = values
        self.created = datetime.now()

        if 'cache-control' in self.values:
            cache_directive = self.values['cache-control']
            max_age = re.findall(r'max-age *= *\d+', cache_directive)
            if len(max_age) > 0:
                cache_seconds = int(max_age[0].split('=')[1])
                self.expires = self.created + timedelta(seconds=cache_seconds)
            else:
                self.expires = None
        else:
            self.expires = None

    @property
    def is_expired(self):
        """Return if the entry is expired or not."""
        return self.expires is not None and datetime.now() > self.expires

    # pylint: disable=invalid-name
    @property
    def st(self):
        """Return ST value."""
        return self.values.get('st')

    @property
    def location(self):
        """Return Location value."""
        return self.values.get('location')

    @property
    def description(self):
        """Return the description from the uPnP entry."""
        url = self.values.get('location', '_NO_LOCATION')

        if url not in UPNPEntry.DESCRIPTION_CACHE:
            try:
                xml = requests.get(url, timeout=5).text
                if not xml:
                    # Samsung Smart TV sometimes returns an empty document the
                    # first time. Retry once.
                    xml = requests.get(url, timeout=5).text

                tree = ElementTree.fromstring(xml)

                UPNPEntry.DESCRIPTION_CACHE[url] = \
                    etree_to_dict(tree).get('root', {})
            except requests.RequestException:
                logging.getLogger(__name__).warning(
                    "Error fetching description at %s", url)

                UPNPEntry.DESCRIPTION_CACHE[url] = {}

            except ElementTree.ParseError:
                logging.getLogger(__name__).warning(
                    "Found malformed XML at %s: %s", url, xml)

                UPNPEntry.DESCRIPTION_CACHE[url] = {}

        return UPNPEntry.DESCRIPTION_CACHE[url]

    def match_device_description(self, values):
        """Fetch description and matches against it.

        Values should only contain lowercase keys.
        """
        device = self.description.get('device')

        if device is None:
            return False

        return all(device.get(key) in val
                   if isinstance(val, list)
                   else val == device.get(key)
                   for key, val in values.items())

    @classmethod
    def from_response(cls, response):
        """Create a uPnP entry from a response."""
        return UPNPEntry({key.lower(): item for key, item
                          in RESPONSE_REGEX.findall(response)})

    def __eq__(self, other):
        """Return the comparison."""
        return (self.__class__ == other.__class__ and
                self.values == other.values)

    def __repr__(self):
        """Return the entry."""
        return "<UPNPEntry {} - {}>".format(
            self.values.get('st', ''), self.values.get('location', ''))


# pylint: disable=invalid-name,too-many-locals
def scan(st=None, timeout=DISCOVER_TIMEOUT, max_entries=None):
    """Send a message over the network to discover uPnP devices.

    Inspired by Crimsdings
    https://github.com/crimsdings/ChromeCast/blob/master/cc_discovery.py

    Protocol explanation:
    https://embeddedinn.wordpress.com/tutorials/upnp-device-architecture/
    """
    # pylint: disable=too-many-nested-blocks,too-many-branches
    ssdp_st = st or ST_ALL
    ssdp_request = "\r\n".join([
        'M-SEARCH * HTTP/1.1',
        'HOST: 239.255.255.250:1900',
        'MAN: "ssdp:discover"',
        'MX: {:d}'.format(SSDP_MX),
        'ST: {}'.format(ssdp_st),
        '', '']).encode('utf-8')

    if st is None:
        # Wemo does not respond to a query for all devices+services
        # but only to a query for just root devices (ST_ROOTDEVICE). Use that
        # as well for the default case.
        alt_ssdp_request = "\r\n".join([
            'M-SEARCH * HTTP/1.1',
            'HOST: 239.255.255.250:1900',
            'MAN: "ssdp:discover"',
            'MX: {:d}'.format(SSDP_MX),
            'ST: {}'.format(ST_ROOTDEVICE),
            '', '']).encode('utf-8')

    stop_wait = datetime.now() + timedelta(0, timeout)

    sockets = []
    for addr in interface_addresses():
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

            # Set the time-to-live for messages for local network
            sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 1)
            sock.bind((addr, 0))

            sockets.append(sock)
        except socket.error:
            pass

    entries = []
    for sock in [s for s in sockets]:
        try:
            sock.sendto(ssdp_request, SSDP_TARGET)
            if alt_ssdp_request:
                sock.sendto(alt_ssdp_request, SSDP_TARGET)
            sock.setblocking(False)
        except socket.error:
            sockets.remove(sock)
            sock.close()

    try:
        while sockets:
            time_diff = stop_wait - datetime.now()
            seconds_left = time_diff.total_seconds()
            if seconds_left <= 0:
                break

            ready = select.select(sockets, [], [], seconds_left)[0]

            for sock in ready:
                try:
                    response = sock.recv(1024).decode("utf-8")
                except socket.error:
                    logging.getLogger(__name__).exception(
                        "Socket error while discovering SSDP devices")
                    sockets.remove(sock)
                    sock.close()
                    continue

                entry = UPNPEntry.from_response(response)

                if (st is None or entry.st == st) and entry not in entries:
                    entries.append(entry)

                    if max_entries and len(entries) == max_entries:
                        raise StopIteration

    except StopIteration:
        pass

    finally:
        for s in sockets:
            s.close()

    return entries


def main():
    """Test SSDP discovery."""
    from pprint import pprint

    pprint("Scanning SSDP..")
    pprint(scan())


if __name__ == "__main__":
    main()



"""Discover devices that implement the Google Cast platform."""
from . import MDNSDiscoverable


# pylint: disable=too-few-public-methods
class Discoverable(MDNSDiscoverable):
    """Add support for discovering Google Cast platform devices."""

    def __init__(self, nd):
        """Initialize the Cast discovery."""
        super(Discoverable, self).__init__(nd, '_googlecast._tcp.local.')



"""Discover Home Assistant iOS app."""
from . import MDNSDiscoverable


# pylint: disable=too-few-public-methods
class Discoverable(MDNSDiscoverable):
    """Add support for discovering the Home Assistant iOS app."""

    def __init__(self, nd):
        super(Discoverable, self).__init__(nd, '_hass-ios._tcp.local.')



"""Discover LG WebOS TV devices."""
from . import SSDPDiscoverable


# pylint: disable=too-few-public-methods
class Discoverable(SSDPDiscoverable):
    """Add support for discovering LG WebOS TV devices."""

    def get_entries(self):
        """Get all the LG WebOS TV device uPnP entries."""
        return self.find_by_device_description(
            {
                "deviceType": "urn:schemas-upnp-org:device:Basic:1",
                "modelName": "LG Smart TV"
            }
        )



"""Discover frontier silicon devices."""
from . import SSDPDiscoverable


class Discoverable(SSDPDiscoverable):
    """Add support for discovering frontier silicon devices."""

    def get_entries(self):
        """Get all the frontier silicon uPnP entries."""
        return [entry for entry in self.netdis.ssdp.all()
                if entry.st and 'fsapi' in entry.st and
                'urn:schemas-frontier-silicon-com' in entry.st]



"""Discover Sonos devices."""
from . import SSDPDiscoverable


# pylint: disable=too-few-public-methods
class Discoverable(SSDPDiscoverable):
    """Add support for discovering Sonos devices."""

    def get_entries(self):
        """Get all the Sonos device uPnP entries."""
        return self.find_by_st("urn:schemas-upnp-org:device:ZonePlayer:1")



"""Add support for discovering mDNS services."""
import zeroconf


class MDNS(object):
    """Base class to discover mDNS services."""

    def __init__(self):
        """Initialize the discovery."""
        self.zeroconf = None
        self.services = []
        self._browsers = []

    def register_service(self, service):
        """Register a mDNS service."""
        self.services.append(service)

    def start(self):
        """Start discovery."""
        try:
            self.zeroconf = zeroconf.Zeroconf()

            for service in self.services:
                self._browsers.append(zeroconf.ServiceBrowser(
                    self.zeroconf, service.typ, service))
        except:  # pylint: disable=broad-except
            self.stop()
            raise

    def stop(self):
        """Stop discovering."""
        while self._browsers:
            self._browsers.pop().cancel()

        for service in self.services:
            service.reset()

        if self.zeroconf:
            self.zeroconf.close()
            self.zeroconf = None

    @property
    def entries(self):
        """Return all entries in the cache."""
        return self.zeroconf.cache.entries()



"""Discover Logitech Media Server."""
from . import BaseDiscoverable


class Discoverable(BaseDiscoverable):
    """Add support for discovering Logitech Media Server."""

    def __init__(self, netdis):
        """Initialize Logitech Media Server discovery."""
        self.netdis = netdis

    def get_entries(self):
        """Get all the Logitech Media Server details."""
        return self.netdis.lms.entries



"""Discover Apple TV media players."""
import ipaddress
from . import MDNSDiscoverable
from ..const import ATTR_HOST, ATTR_NAME


# pylint: disable=too-few-public-methods
class Discoverable(MDNSDiscoverable):
    """Add support for Apple TV devices."""

    def __init__(self, nd):
        super(Discoverable, self).__init__(nd, '_appletv-v2._tcp.local.')

    def info_from_entry(self, entry):
        """Returns most important info from mDNS entries."""
        props = entry.properties
        return {
            ATTR_HOST: str(ipaddress.ip_address(entry.address)),
            ATTR_NAME: props.get(b'Name').decode('utf-8').replace('\xa0', ' '),
            'hsgid': props.get(b'hG').decode('utf-8')
            }



"""Util functions used by Netdisco."""
from collections import defaultdict

import netifaces


# Taken from http://stackoverflow.com/a/10077069
# pylint: disable=invalid-name
def etree_to_dict(t):
    """Convert an ETree object to a dict."""
    # strip namespace
    tag_name = t.tag[t.tag.find("}")+1:]

    d = {tag_name: {} if t.attrib else None}
    children = list(t)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d = {tag_name: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if t.attrib:
        d[tag_name].update(('@' + k, v) for k, v in t.attrib.items())
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            if text:
                d[tag_name]['#text'] = text
        else:
            d[tag_name] = text
    return d


def interface_addresses(family=netifaces.AF_INET):
    """Return local addresses of any associated network.

    Gathering of addresses which are bound to a local interface that has
    broadcast (and probably multicast) capability.
    """
    # pylint: disable=no-member
    return [addr['addr']
            for i in netifaces.interfaces()
            for addr in netifaces.ifaddresses(i).get(family) or []
            if 'broadcast' in addr]



"""Discover Yamaha Receivers."""
from . import SSDPDiscoverable


class Discoverable(SSDPDiscoverable):
    """Add support for discovering Yamaha Receivers."""

    REMOTE_CONTROL_SPEC_TYPE =\
        'urn:schemas-yamaha-com:service:X_YamahaRemoteControl:1'

    def info_from_entry(self, entry):
        """Return the most important info from a uPnP entry."""
        info = super().info_from_entry(entry)

        yam = entry.description['X_device']
        services = yam['X_serviceList']['X_service']
        if isinstance(services, list):
            service = next(
                (s for s in services
                 if s['X_specType'] == self.REMOTE_CONTROL_SPEC_TYPE),
                services[0])
        else:
            service = services
        # do a slice of the second element so we don't have double /
        info['control_url'] = yam['X_URLBase'] + service['X_controlURL'][1:]
        info['description_url'] = (yam['X_URLBase'] +
                                   service['X_unitDescURL'][1:])

        return info

    def get_entries(self):
        """Get all the Yamaha uPnP entries."""
        return self.find_by_device_description({
            "manufacturer": "Yamaha Corporation",
            "deviceType": "urn:schemas-upnp-org:device:MediaRenderer:1"
        })



from setuptools import setup, find_packages

setup(name='netdisco',
      version='1.0.0.rc1',
      description='Discover devices on your local network',
      url='https://github.com/home-assistant/netdisco',
      author='Paulus Schoutsen',
      author_email='Paulus@PaulusSchoutsen.nl',
      license='Apache License 2.0',
      install_requires=['requests>=2.0', 'zeroconf==0.19'],
      packages=find_packages(exclude=['tests', 'tests.*']),
      zip_safe=False)



"""Discover Yeelight bulbs, based on Kodi discoverable."""
import logging
from . import MDNSDiscoverable
from ..const import ATTR_DEVICE_TYPE


# pylint: disable=too-few-public-methods
class Discoverable(MDNSDiscoverable):
    """Add support for discovering Yeelight."""

    def __init__(self, nd):
        """Initialize the Yeelight discovery."""
        super(Discoverable, self).__init__(nd, '_miio._udp.local.')

    def info_from_entry(self, entry):
        """Return most important info from mDNS entries."""
        info = super().info_from_entry(entry)

        device_type = "UNKNOWN"
        if entry.name.startswith("yeelink-light-color1_"):
            device_type = "rgb"
        elif entry.name.startswith("yeelink-light-mono1_"):
            device_type = "white"
        else:
            logging.warning("Unknown miio device found: %s", entry)

        info[ATTR_DEVICE_TYPE] = device_type
        return info

    def get_entries(self):
        """ Return yeelight devices. """
        return self.find_by_device_name('yeelink-light-')



"""Discover Denon AVR devices."""
from urllib.parse import urlparse

from . import SSDPDiscoverable
from ..const import ATTR_HOST


class Discoverable(SSDPDiscoverable):
    """Add support for discovering Denon AVR devices."""

    def get_entries(self):
        """Get all Denon AVR uPnP entries."""
        return self.find_by_device_description({
            "manufacturer": "Denon",
            "deviceType": "urn:schemas-upnp-org:device:MediaRenderer:1"
        })

    def info_from_entry(self, entry):
        """Get most important info, which is name, model and host."""
        info = super().info_from_entry(entry)
        info[ATTR_HOST] = urlparse(
            entry.description['device']['presentationURL']).hostname
        return info



"""Constants of services that can be discovered."""

BELKIN_WEMO = "belkin_wemo"
DLNA = "DLNA"
GOOGLE_CAST = "google_cast"
PHILIPS_HUE = "philips_hue"
PMS = 'plex_mediaserver'
LMS = 'logitech_mediaserver'
NETGEAR_ROUTER = "netgear_router"
SONOS = "sonos"
PANASONIC_VIERA = "panasonic_viera"
SABNZBD = 'sabnzbd'
KODI = 'kodi'
HOME_ASSISTANT = "home_assistant"
MYSTROM = 'mystrom'
HASS_IOS = "hass_ios"
BOSE_SOUNDTOUCH = 'bose_soundtouch'
SAMSUNG_TV = "samsung_tv"
FRONTIER_SILICON = "frontier_silicon"
APPLE_TV = "apple_tv"
HARMONY = "harmony"

ATTR_NAME = 'name'
ATTR_HOST = 'host'
ATTR_PORT = 'port'
ATTR_HOSTNAME = 'hostname'
ATTR_URLBASE = 'urlbase'
ATTR_DEVICE_TYPE = 'device_type'
ATTR_MODEL_NAME = 'model_name'
ATTR_MODEL_NUMBER = 'model_number'
ATTR_PROPERTIES = 'properties'
ATTR_SSDP_DESCRIPTION = 'ssdp_description'
ATTR_SERIAL = 'serial'
ATTR_MAC_ADDRESS = 'mac_address'



"""Discover Panasonic Viera TV devices."""
from . import SSDPDiscoverable


# pylint: disable=too-few-public-methods
class Discoverable(SSDPDiscoverable):
    """Add support for discovering Viera TV devices."""

    def get_entries(self):
        """Get all the Viera TV device uPnP entries."""
        return self.find_by_st("urn:panasonic-com:service:p00NetworkControl:1")



"""Provides helpful stuff for discoverables."""
# pylint: disable=abstract-method
from urllib.parse import urlparse

from ..const import (
    ATTR_NAME, ATTR_MODEL_NAME, ATTR_HOST, ATTR_PORT, ATTR_SSDP_DESCRIPTION,
    ATTR_SERIAL, ATTR_MODEL_NUMBER, ATTR_HOSTNAME, ATTR_PROPERTIES)


class BaseDiscoverable(object):
    """Base class for discoverable services or device types."""

    def is_discovered(self):
        """Return True if it is discovered."""
        return len(self.get_entries()) > 0

    def get_info(self):
        """Return a list with the important info for each item.

        Uses self.info_from_entry internally.
        """
        return [self.info_from_entry(entry) for entry in self.get_entries()]

    # pylint: disable=no-self-use
    def info_from_entry(self, entry):
        """Return an object with important info from the entry."""
        return entry

    # pylint: disable=no-self-use
    def get_entries(self):
        """Return all the discovered entries."""
        raise NotImplementedError()


class SSDPDiscoverable(BaseDiscoverable):
    """uPnP discoverable base class."""

    def __init__(self, netdis):
        """Initialize SSDPDiscoverable."""
        self.netdis = netdis

    def info_from_entry(self, entry):
        """Get most important info, by default the description location."""
        url = urlparse(entry.location)
        info = {
            ATTR_HOST: url.hostname,
            ATTR_PORT: url.port,
            ATTR_SSDP_DESCRIPTION: entry.location
        }
        device = entry.description.get('device')

        if device:
            info[ATTR_NAME] = device.get('friendlyName')
            info[ATTR_MODEL_NAME] = device.get('modelName')
            info[ATTR_MODEL_NUMBER] = device.get('modelNumber')
            info[ATTR_SERIAL] = device.get('serialNumber')

        return info

    # Helper functions

    # pylint: disable=invalid-name
    def find_by_st(self, st):
        """Find entries by ST (the device identifier)."""
        return self.netdis.ssdp.find_by_st(st)

    def find_by_device_description(self, values):
        """Find entries based on values from their description."""
        return self.netdis.ssdp.find_by_device_description(values)


class MDNSDiscoverable(BaseDiscoverable):
    """mDNS Discoverable base class."""

    def __init__(self, netdis, typ):
        """Initialize MDNSDiscoverable."""
        self.netdis = netdis
        self.typ = typ
        self.services = {}

        netdis.mdns.register_service(self)

    def reset(self):
        """Reset found services."""
        self.services.clear()

    def is_discovered(self):
        """Return True if any device has been discovered."""
        return len(self.get_entries()) > 0

    # pylint: disable=unused-argument
    def remove_service(self, zconf, typ, name):
        """Callback when a service is removed."""
        self.services.pop(name, None)

    def add_service(self, zconf, typ, name):
        """Callback when a service is found."""
        service = None
        tries = 0
        while service is None and tries < 3:
            service = zconf.get_service_info(typ, name)
            tries += 1

        if service is not None:
            self.services[name] = service

    def get_entries(self):
        """Return all found services."""
        return self.services.values()

    def info_from_entry(self, entry):
        """Return most important info from mDNS entries."""
        properties = {}

        for key, value in entry.properties.items():
            if isinstance(value, bytes):
                value = value.decode('utf-8')
            properties[key.decode('utf-8')] = value

        return {
            ATTR_HOST: self.ip_from_host(entry.server),
            ATTR_PORT: entry.port,
            ATTR_HOSTNAME: entry.server,
            ATTR_PROPERTIES: properties,
        }

    def find_by_device_name(self, name):
        """Find entries based on the beginning of their entry names."""
        return [entry for entry in self.services.values()
                if entry.name.startswith(name)]

    def ip_from_host(self, host):
        """Attempt to return the ip address from an mDNS host.

        Return host if failed.
        """
        ips = self.netdis.mdns.zeroconf.cache.entries_with_name(host.lower())

        try:
            return repr(ips[0]) if ips else host
        except TypeError:
            return host


class GDMDiscoverable(BaseDiscoverable):
    """GDM discoverable base class."""

    def __init__(self, netdis):
        """Initialize GDMDiscoverable."""
        self.netdis = netdis

    def info_from_entry(self, entry):
        """Get most important info, by default the description location."""
        return {
            ATTR_HOST: entry.values['location'],
            ATTR_PORT: entry.values['port'],
        }

    def find_by_content_type(self, value):
        """Find entries based on values from their content_type."""
        return self.netdis.gdm.find_by_content_type(value)

    def find_by_data(self, values):
        """Find entries based on values from any returned field."""
        return self.netdis.gdm.find_by_data(values)
